"""
Job commands
"""
import base64
import glob
import json
import os
import shutil
import stat
import subprocess
import sys
import threading
from distutils import dir_util

from onepanel.utilities.aws_utility import AWSUtility

import click
import configobj
import websocket

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import ProjectViewController
from onepanel.utilities.s3_authenticator import S3Authenticator


class JobViewController(APIViewController):

    JOB_OUTPUT_FILE = os.path.join('.onepanel','output')

    project_account_uid = None
    project_uid = None
    s3_bucket_name = 'onepanel-datasets'

    def __init__(self, conn):
        APIViewController.__init__(self, conn)

    def init_credentials_retrieval(self):
        # Figure out the  account uid and project uid from file
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        if not os.path.isfile(project_file):
            print("ERROR.Project file does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + project_file)
            exit(-1)

        cfg = configobj.ConfigObj(project_file)

        project_uid = cfg['uid']
        project_account_uid = cfg['account_uid']

        if len(project_uid) < 1 or len(project_account_uid) < 1:
            print("ERROR.Project file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.project_account_uid = project_account_uid
        self.project_uid = project_uid

    def init_credentials_job_output(self):
        # Figure out the job uid, account uid and project uid from file
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        job_output_file = os.path.join(home, JobViewController.JOB_OUTPUT_FILE)
        if not os.path.isfile(job_output_file):
            print("ERROR.Job file does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + job_output_file)
            exit(-1)

        cfg = configobj.ConfigObj(job_output_file)

        job_uid = cfg['uid']
        project_uid = cfg['project_uid']
        account_uid = cfg['account_uid']

        if len(job_uid) < 1 or len(project_uid) < 1 or len(account_uid) < 1:
            print("ERROR.Project file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.job_uid = job_uid
        self.project_uid = project_uid
        self.project_account_uid = account_uid

    def init_endpoint(self):
        # Generate the endpoint
        self.endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/jobs'.format(
            root=self.conn.URL,
            account_uid=self.project_account_uid,
            project_uid=self.project_uid
        )

    @staticmethod
    def get_s3_root_for_job_output(account_uid='', project_uid='', job_uid=''):
        return '{account_uid}/projects/{project_uid}/jobs/{job_uid}/output'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            job_uid=job_uid
        )

    @staticmethod
    def get_s3_compressed_file_for_job_output(account_uid='',project_uid='',job_uid=''):
        return '{account_uid}/projects/{project_uid}/jobs/{job_uid}/job-{job_uid}-output.tar.gz'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            job_uid=job_uid
        )


@click.group(help='Job commands group')
@click.pass_context
def jobs(ctx):
    ctx.obj['vc'] = JobViewController(ctx.obj['connection'])
    project_vc = ProjectViewController(ctx.obj['connection'])
    path = os.getcwd()
    if project_vc.exists_local(path):
        project_vc.from_directory(path)
        ctx.obj['project'] = project_vc
    else:
        # Return error before reaching other CLI commands:
        print('This project is not initialized, type "onepanel init" to initialize this project')
        sys.exit(1)


@jobs.command('create', help='Execute a command on a remote machine in the current project')
@click.argument(
    'command',
    type=str
)
@click.option(
    '-m', '--machine-type',
    type=str,
    required=True,
    help='Machine type ID'
)
@click.option(
    '-e', '--environment',
    type=str,
    required=True,
    help='Instance template ID'
)
@click.pass_context
@login_required
def create_job(ctx, command, machine_type, environment):
    new_job = {
        'command': command,
        'machineType': {
            'uid': machine_type
        },
        'instanceTemplate': {
            'uid': environment
        }
    }

    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    response = vc.post(post_object=new_job)

    if response['status_code'] == 200:
        print('New job created: {}'.format(response['data']['uid']))
    else:
        print("An error occurred: {}".format(response['data']))

    return


@jobs.command('list', help='Show commands executed on remote machines')
@click.option(
    '-a', '--all',
    type=bool,
    is_flag=True,
    default=False,
    help='Include finished commands'
)
@click.pass_context
@login_required
def list_jobs(ctx, all):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    items = vc.list(params='?running=true' if not all else '')

    if items is None or len(items) == 0:
        print('No jobs found. Use "--all" flag to retrieve completed jobs.')
        return

    vc.print_items(items, fields=['uid', 'command'], field_names=['ID', 'COMMAND'])


@jobs.command('stop', help='Stop a job')
@click.argument(
    'job_uid',
    type=str
)
@click.pass_context
@login_required
def kill_job(ctx, job_uid):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    ctx.obj['vc'].delete(job_uid, field_path='/active', message_on_success='Job stopped',
                         message_on_failure='Job not found')


@jobs.command('logs', help='Show a log of the command')
@click.argument(
    'job_uid',
    type=str
)
@click.pass_context
@login_required
def job_logs(ctx, job_uid):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    project = ctx.obj['project']

    job_data = vc.get('/' + job_uid, field_path='/logs')

    if job_data['status_code'] != 400:
        if job_data['data'] is None:
            print('Job not found.')
            return
        log = job_data['data']
        print(log)
    else:
        # Streaming via WebSocket
        # See https://pypi.python.org/pypi/websocket-client/

        def on_message(ws, message):
            if message[0] == '0':
                message = base64.b64decode(message[1:]).decode('utf-8', 'ignore')
                sys.stdout.write(message)
                sys.stdout.flush()

        def on_error(ws, error):
            if isinstance(error, websocket.WebSocketConnectionClosedException):
                return

            if isinstance(error, KeyboardInterrupt):
                return

            if error.status_code == 502 or error.status_code == 503:
                print('Job {} is preparing'.format(job_uid))
            else:
                print(error)

        def on_close(ws):
            print('connection closed')

        def on_open(ws):
            def send_auth_token(*args):
                ws.send(json.dumps({'Authtoken': ''}))

            threading.Thread(target=send_auth_token).start()

        ws_url = '{ws_root}/{account_uid}/projects/{project_uid}/jobs/{job_uid}/logs/ws?id_token={token}'.format(
            ws_root='wss://c.onepanel.io',
            account_uid=project.account_uid,
            project_uid=project.project_uid,
            job_uid=job_uid,
            token=ctx.obj['connection'].token
        )

        ws = websocket.WebSocketApp(
            url=ws_url,
            on_message=on_message,
            on_open=on_open,
            on_error=on_error
        )

        ws.run_forever()

    return False


def jobs_download_output(ctx, path, directory,archive_flag):
    jvc = JobViewController(ctx.obj['connection'])
    #
    # Resource
    #
    dirs = path.split('/')
    git_utility = ctx.obj['git_utility']

    # Job output: Method 2
    # <account_uid>/projects/<project_uid>/jobs/<job_uid>
    if len(dirs) == 5:
        try:
            account_uid, projects_dir, project_uid, jobs_dir, job_uid = dirs
            assert (projects_dir == 'projects') and (jobs_dir == 'jobs')
        except:
            print('Incorrect job path')
            return None
    else:
        print('Incorrect job uid')
        return None

    #
    # Directory
    #
    if directory is None or directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    # Check how the job output is stored
    jvc.project_account_uid = account_uid
    jvc.project_uid = project_uid
    jvc.init_endpoint()
    job_data = jvc.get('/' + job_uid)
    if job_data['data'] is None:
        print("Job not found.")
        return False
    is_legacy = job_data['data']['isLegacy']
    if is_legacy:
        if archive_flag is True:
            print("--archive is not supported for this job output.")
        else:
            #
            # Clone
            #
            onepanel_download_path = '.onepanel_download'
            if os.path.isdir(onepanel_download_path):
                shutil.rmtree(onepanel_download_path,
                              onerror=remove_readonly)  # Cleaning after previous errors to avoid "is not an empty directory" error
            cmd = (
                    git_utility.get_git_clone_str() + ' -b job-{job_uid} https://{user_uid}:{gitlab_token}@{host}/{account_uid}/{project_uid}-output.git '
                                                      '.onepanel_download'
            ).format(
                host=os.getenv('GITLAB_GIT_HOST', 'git.onepanel.io'),
                user_uid=ctx.obj['connection'].user_uid,
                gitlab_token=ctx.obj['connection'].gitlab_impersonation_token,
                account_uid=account_uid,
                project_uid=project_uid,
                job_uid=job_uid,
                dir=home
            )
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            p.wait()
            status_code = p.returncode
            if status_code == 128:
                print("Could not find the remote git repository.")
                exit(0)

            jobs_path = '/jobs/{job_uid}/output/'.format(job_uid=job_uid)
            copy_from_dir = onepanel_download_path + jobs_path
            if not os.path.exists(copy_from_dir):
                print("This job did not create any output or output was not saved. \n" +
                      "If you want to save and version control your output, modify your script to "
                      "save all output to the '/onepanel/output' directory.\n")
                if os.path.isdir(onepanel_download_path):
                    shutil.rmtree(onepanel_download_path,
                                  onerror=remove_readonly)
                exit(0)
            else:
                dir_util.copy_tree(copy_from_dir, home)

            if os.path.isdir(onepanel_download_path):
                shutil.rmtree(onepanel_download_path, onerror=remove_readonly)

            if p.returncode == 0:
                print('Job output downloaded to: {dir}'.format(dir=home))
                return True
            else:
                print('Unable to download')
                return False
    else:
        s3auth = S3Authenticator(ctx.obj['connection'])
        creds = s3auth.get_s3_credentials(jvc.project_account_uid, 'projects', jvc.project_uid)
        if creds['data'] is None:
            print("Unable to get S3 credentials. Exiting.")
            return False
        aws_access_key_id = creds['data']['AccessKeyID']
        aws_secret_access_key = creds['data']['SecretAccessKey']
        aws_session_token = creds['data']['SessionToken']
        aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                              aws_secret_access_key=aws_secret_access_key,
                              aws_session_token=aws_session_token)
        s3_download_to_path = home
        if archive_flag is True:
            print("Attempting to download the compressed output file to {home} directory.".format(home=s3_download_to_path))
            s3_path_to_download_from = jvc.get_s3_compressed_file_for_job_output(account_uid, project_uid, job_uid)
            full_path = 's3://{bucket}/{path}'.format(bucket=aws_util.get_dataset_bucket_name(),
                                                      path=s3_path_to_download_from)
            investigation_results = aws_util.check_s3_path_for_files(full_path,False)
            if investigation_results['code'] == -1:
                print("Error encountered.")
                print(investigation_results['msg'])
                return
            if investigation_results['code'] == 0 and investigation_results['data'] == 0:
                print("This job did not create any output or output was not saved. \n" +
                      "If you want to save and version control your output, modify your script to "
                      "save all output to the '/onepanel/output' directory.\n")
                return
            exit_code = aws_util.download(s3_download_to_path,s3_path_to_download_from)
            if exit_code != 0:
                print("Error encountered.")
                return
        else:
            print("Attempting to download output to {home} directory.".format(home=s3_download_to_path))
            s3_path_to_download_from = jvc.get_s3_root_for_job_output(account_uid, project_uid, job_uid)
            full_path = 's3://{bucket}/{path}'.format(bucket=aws_util.get_dataset_bucket_name(),path=s3_path_to_download_from)
            investigation_results = aws_util.check_s3_path_for_files(full_path)
            if investigation_results['code'] == -1:
                print("Error encountered.")
                print(investigation_results['msg'])
                return
            if investigation_results['code'] == 0 and investigation_results['data'] == 0:
                print("This job did not create any output or output was not saved. \n" +
                    "If you want to save and version control your output, modify your script to "
                    "save all output to the '/onepanel/output' directory.\n")
                return
            # Check if there any actual files to download from the output
            exit_code = aws_util.download_all(s3_download_to_path,s3_path_to_download_from)
            if exit_code != 0:
                print("Error encountered.")
                return
        print("Finished downloading.")



@jobs.command('delete', help='Delete a job')
@click.argument(
    'job_uid',
    type=str
)
@click.pass_context
@login_required
def delete_job(ctx, job_uid):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    ctx.obj['vc'].delete(job_uid, message_on_success='Job deleted', message_on_failure='Job not found')


def remove_readonly(func, path, excinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)

def upload_output(ctx,suppress_output=False):
    ctx.obj['vc'] = JobViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.init_credentials_job_output()

    files = glob.glob("*")
    if len(files) < 1:
        click.echo("Cannot find any files in current dir, exiting.")
        return

    # Authenticate for S3
    s3auth = S3Authenticator(ctx.obj['connection'])
    creds = s3auth.get_s3_credentials(vc.project_account_uid, 'projects', vc.project_uid)
    if creds['data'] is None:
        print("Unable to get S3 credentials. Exiting.")
        return False
    aws_access_key_id = creds['data']['AccessKeyID']
    aws_secret_access_key = creds['data']['SecretAccessKey']
    aws_session_token = creds['data']['SessionToken']

    job_source_dir = os.curdir
    s3_push_to_dir = '{account_uid}/projects/{project_uid}/jobs/{job_uid}'.format(
        account_uid=vc.project_account_uid,
        project_uid=vc.project_uid,
        job_uid=vc.job_uid
    )

    aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                          aws_secret_access_key=aws_secret_access_key,
                          aws_session_token=aws_session_token)
    aws_util.suppress_output = suppress_output

    exit_code = aws_util.upload_dir(job_source_dir, s3_push_to_dir, '.onepanel/*')
    if exit_code != 0:
        click.echo('\nError with pushing up files.')
        return
    print("Finished output upload.")


import click

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required


class VolumeTypeViewController(APIViewController):

    def __init__(self, conn):
        APIViewController.__init__(self,conn)


@click.group(name='volume-types', help='Available volume types')
@click.pass_context
def volume_types(ctx):
    ctx.obj['view_controller'] = VolumeTypeViewController(ctx.obj['connection'])


@volume_types.command(
    'list',
    help='Show available volume types and their IDs'
)
@click.pass_context
@login_required
def list_volume_types(ctx):

    vc = ctx.obj['view_controller']

    items = vc.list(params='/volume_types')
    vc.print_items(items, fields=['uid', 'name'], field_names=['ID', 'SPECS'])




"""
Machine types commands
"""

import click

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required


class EnvironmentViewController(APIViewController):

    def __init__(self,conn):
        APIViewController.__init__(self,conn)

@click.group(help='Environment (machine types) commands group')
@click.pass_context
def environments(ctx):
    ctx.obj['vc'] = EnvironmentViewController(ctx.obj['connection'])


@environments.command('list', help='Show available environments')
@click.pass_context
@login_required
def list_environments(ctx):
    vc = ctx.obj['vc']
    print_data = vc.list(params='/instance_templates?instance_type=mod')
    fields = ['uid','name']
    field_names = ['ID','ENVIRONMENT']
    empty_message = 'No environments found.'
    vc.print_items(print_data,fields,field_names,empty_message)




import onepanel.git_hooks.pre_push_to_endpoint_hook

class PrePush:
    version = None

    def __init__(self):
        self.version = '1.0.0'

    @staticmethod
    def read_self():
        file_path = __file__
        if file_path.endswith('.pyc'):
            file_path = file_path.replace('pyc','py')
        with open(file_path,'r') as f:
            return f.read()


if __name__ == '__main__':
    onepanel.git_hooks.pre_push_to_endpoint_hook.PrePushToEndpointHook.ping_endpoint()


import platform
import subprocess
import os
import sys
from distutils.sysconfig import get_python_lib


class AWSUtility:
	env = {} # Use for shell command environment variables
	suppress_output = False

	def __init__(self,aws_access_key_id='',
				 aws_secret_access_key='',
				 aws_session_token=''):
		self.aws_access_key_id = aws_access_key_id
		self.aws_secret_access_key = aws_secret_access_key
		self.aws_session_token = aws_session_token
		self.env = {'AWS_ACCESS_KEY_ID': self.aws_access_key_id, 'AWS_SECRET_ACCESS_KEY': self.aws_secret_access_key,
			   'AWS_SESSION_TOKEN': self.aws_session_token,
		}
		if platform.system() is 'Windows':
			self.env[str('SYSTEMROOT')] = os.environ['SYSTEMROOT']
			self.env[str('PATH')] = os.environ['PATH']
			self.env[str('PYTHONPATH')] = os.pathsep.join(sys.path)


	def get_dataset_bucket_name(self):
		return os.getenv('DATASET_BUCKET', 'onepanel-datasets')

	def upload_dir(self, dataset_directory, s3_directory,exclude=''):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			exclude_arg = '--exclude "{ex_str}"'.format(ex_str=exclude)
			cmd = ' '.join([aws_full_path, 's3', 'sync', dataset_directory, s3_path,exclude_arg])
			# Also need to execute through the shell
			print("Uploading...")
			p = subprocess.Popen(args=cmd, env=self.env,stdin=subprocess.PIPE, stdout=subprocess.PIPE,
						 		stderr=subprocess.PIPE, shell=True)
			suppress_s3_info = s3_path.rsplit('/', 1)[0]
			for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
				if self.suppress_output is False:
					line_str = line.decode()
					cleaned_line = line_str.replace(suppress_s3_info, '')
					sys.stdout.write(cleaned_line)
			for line in iter(p.stderr.readline, '' or b''):  # replace '' with b'' for Python 3
				if self.suppress_output is False:
					line_str = line.decode()
					cleaned_line = line_str.replace(suppress_s3_info, '')
					sys.stdout.write(cleaned_line)
			return 0
		return -1

	def download_all(self, dataset_directory, s3_directory):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join([aws_full_path, 's3', 'sync', s3_path, dataset_directory])
			print("Downloading...")
			# Also need to execute through the shell
			p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=True)
			suppress_s3_info = s3_path.rsplit('/', 1)[0]
			for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
				if self.suppress_output is False:
					line_str = line.decode()
					cleaned_line = line_str.replace(suppress_s3_info, '')
					sys.stdout.write(cleaned_line)
			return 0
		return -1

	def download(self,to_dir,s3_full_path_to_file):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_full_path_to_file)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join([aws_full_path,'s3', 'cp', s3_path, to_dir])
			print("Downloading...")
			# Also need to execute through the shell
			p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=True)
			suppress_s3_info = s3_path.rsplit('/',1)[0]
			for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
				if self.suppress_output is False:
					line_str = line.decode()
					cleaned_line = line_str.replace(suppress_s3_info,'')
					sys.stdout.write(cleaned_line)
			return 0
		return -1

	def check_s3_path_for_files(self,full_s3_path='',recursive=True):
		ret_val = {'data':None, 'code':-1, 'msg':''}
		if full_s3_path == '':
			ret_val = {'data': None, 'code': -1, 'msg': 'Need the full s3 path passed in.'}
			return ret_val
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			recursive_arg = ''
			if recursive:
				recursive_arg = '--recursive'
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join([aws_full_path, 's3', 'ls',recursive_arg,'--summarize',full_s3_path])
			# Also need to execute through the shell
			p = subprocess.Popen(args=cmd, env=self.env,stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=True)
			p.wait()
			raw_output = p.stdout.readlines()
			for line in raw_output:
				if 'Total Objects' in line.decode():
					objects_str = line.decode()
					str_split = objects_str.split(':')
					# Grab the string of the number
					string_num = str_split[-1].strip()
					ret_val = {'data': int(string_num), 'code': 0, 'msg': 'Total files found.'}
					break
			return ret_val
		return ret_val

	def get_full_path_to_aws_cli(self):
		# Figure out the full path to awscli
		if platform.system() is 'Windows':
			path_to_aws_cmd = ['where','aws']
		else:
			path_to_aws_cmd = ['which', 'aws']
		p = subprocess.Popen(args=path_to_aws_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=False)
		p.wait()
		line = p.stdout.readline()
		return line.decode().rstrip()



import os
import functools
import six
if six.PY2:
    import ConfigParser as configparser
else:
    import configparser
import requests

class Connection:
    """ REST API requests defaults and credentials
    """
    def __init__(self):
        self.URL = os.getenv('BASE_API_URL', 'https://c.onepanel.io/api')
        self.SSL_VERIFY = True
        self.headers = {'Content-Type': 'application/json'}
        self.account_uid = None
        self.token = None
        self.gitlab_impersonation_token = None

        # wrap requests methods to reduce number of arguments in api queries
        self.get = functools.partial(requests.get, headers=self.headers, verify=self.SSL_VERIFY)
        self.post = functools.partial(requests.post, headers=self.headers, verify=self.SSL_VERIFY)
        self.put = functools.partial(requests.put, headers=self.headers, verify=self.SSL_VERIFY)
        self.delete = functools.partial(requests.delete, headers=self.headers, verify=self.SSL_VERIFY)
        self.head = functools.partial(requests.head, headers=self.headers, verify=self.SSL_VERIFY)

    def save_credentials(self, data):
        credentials = configparser.ConfigParser()
        credentials.add_section('onepanel')
        credentials.set('onepanel','uid',data['uid'])
        credentials.set('onepanel','token',data['sessions'][0]['token'])
        credentials.set('onepanel','gitlab_impersonation_token',data['gitlab_impersonation_token'])
        credentials.set('onepanel','account_uid',data['account']['uid'])

        onepanel_home = os.path.expanduser(os.path.join('~', '.onepanel'))
        if not os.path.exists(onepanel_home):
            os.makedirs(onepanel_home)

        filename = os.path.join(onepanel_home, 'credentials')
        with open(filename, 'w') as f:
            credentials.write(f)

    def load_credentials(self):
        credentials = configparser.ConfigParser()
        filename = os.path.expanduser(os.path.join('~', '.onepanel', 'credentials'))
        credentials.read(filename)

        try:
            self.account_uid = credentials.get('onepanel', 'account_uid')
        except configparser.NoSectionError as e:
            self.account_uid = None

        try:
            self.token = credentials.get('onepanel', 'token')
        except configparser.NoSectionError as e:
            self.token = None

        try:
            self.gitlab_impersonation_token = credentials.get('onepanel', 'gitlab_impersonation_token')
        except configparser.NoSectionError as e:
            self.gitlab_impersonation_token = None

        if self.token:
            self.headers['Authorization'] = 'Bearer {}'.format(self.token)




"""
View-Controller class for OnePanel API
"""
import json

from prettytable import PrettyTable


class APIViewController:
    conn = None  # Connection from onepanel.cli
    endpoint = str

    def __init__(self, conn):

        self.conn = conn
        self.endpoint = conn.URL

    # Wrappers for HTTP requests
    def get(self, uid='', field_path='', params=None):
        """Get a JSON object from the endpoint"""

        r = self.conn.get(
            '{endpoint}{uid}{field_path}{params}'.format(endpoint=self.endpoint, uid=uid, field_path=field_path,
                                                          params=params or ''))

        c = r.status_code
        if c == 200:
            if r.headers['Content-Type'] == 'text/plain':
                item = r.text
            else:
                item = r.json()
        else:
            item = None

        response_data = {
                'data':item,
                'status_code':c
            }
        return response_data

    def put(self, uid='', field_path='', params=None,post_object=None):
        """Put a JSON object to the endpoint"""

        url = '{endpoint}{uid}{field_path}{params}'.format(endpoint=self.endpoint, uid=uid, field_path=field_path,
                                                         params=params or '')
        r = self.conn.put(url,data=json.dumps(post_object))

        c = r.status_code
        if c == 200:
            if r.headers['Content-Type'] == 'text/plain':
                item = r.text
            else:
                item = r.json()
        else:
            item = None

        response_data = {
            'data': item,
            'status_code': c
        }
        return response_data

    def head(self, uid='', field_path='', params=None):
        """Make a head request to the endpoint"""

        url = '{endpoint}{uid}{field_path}{params}'.format(endpoint=self.endpoint, uid=uid, field_path=field_path,
                                                           params=params or '')
        r = self.conn.head(url)

        c = r.status_code
        if c == 200:
            if r.headers['Content-Type'] == 'text/plain':
                item = r.text
            else:
                item = r.json()
        else:
            item = None

        response_data = {
            'data': item,
            'status_code': c
        }
        return response_data

    def list(self, params=None):
        """Get a JSON list from the endpoint"""

        r = self.conn.get('{endpoint}{params}'.format(endpoint=self.endpoint, params=params or ''))

        c = r.status_code
        if c == 200:
            items = r.json()
        else:
            return None

        return items

    def post(self,post_object,params=None):
        """POST an object and return the newly created resource"""

        url = '{endpoint}{params}'.format(endpoint=self.endpoint, params=params or '')
        r = self.conn.post(url, data=json.dumps(post_object))

        c = r.status_code
        if c == 200:
            created_object = r.json()
        else:
            try:
                created_object = "An error occurred: {}".format(r.json())
            except ValueError:
                created_object = "An error occurred: {}".format(r.text)

        return {
            'data': created_object,
            'status_code': c
        }

    def delete(self, uid, field_path='', message_on_success='Resource deleted',
               message_on_failure='Resource not found'):
        r = self.conn.delete('{}/{}{}'.format(self.endpoint, uid, field_path))

        c = r.status_code
        if c == 200:
            print('{message}: {id}'.format(message=message_on_success, id=uid))
            return True
        elif c == 404:
            print('{message}: {id}'.format(message=message_on_failure, id=uid))
        else:
            print('Error: {}'.format(c))

        return False

    # Support functions

    @staticmethod
    def print_items(
            items,
            fields=None,
            field_names=None,
            empty_message='No items found'
    ):
        """A standard table for showing response to GET requests"""

        if fields is None:
            fields = ['uid', 'name']
        if field_names is None:
            field_names = ['ID', 'NAME']

        if len(items) == 0:
            print(empty_message)
            return items

        summary = PrettyTable(border=False)
        summary.field_names = field_names
        summary.align = 'l'
        for item in items:
            summary.add_row([item[field] for field in fields])
        print(summary)



"""
Workspace module
"""
import os
import re
import sys

import click
import configobj

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import ProjectViewController


class InstanceViewController(APIViewController):

    account_uid = None
    uid = None

    def __init__(self, conn):
        APIViewController.__init__(self, conn)


    def init_credentials_retrieval(self):
        home = os.getcwd()
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)

        if len(cfg) == 0:
            print("ERROR.Cannot figure out the current project.")
            print("Make sure .onepanel/project exists and is accurate.")
            exit(-1)
        self.account_uid = cfg['account_uid']
        self.uid = cfg['uid']

    def init_endpoint(self):
        self.endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=self.account_uid,
            project_uid=self.uid
        )

@click.group(help='Workspace commands group')
@click.pass_context
def workspaces(ctx):
    ctx.obj['vc'] = InstanceViewController(ctx.obj['connection'])


@workspaces.command(
    'create',
    help='Create a new workspace. The workspace\'s name: Max 25 chars, lower case alphanumeric or "-", '
         'must start and end with alphanumeric'
)
@click.argument(
    'workspace_uid',
    type=str
)
@click.option(
    '-m', '--machine-type',
    type=str,
    required=True,
    help='Machine type ID. Call "onepanel machine_types list" for IDs.'
)
@click.option(
    '-e', '--environment',
    type=str,
    required=True,
    help='Instance template ID. Call "onepanel environments list" for IDs.'
)
@click.option(
    '-s', '--storage',
    type=str,
    required=True,
    help='Storage type ID.'
)
@click.pass_context
@login_required
def create_instance(ctx, workspace_uid, machine_type, environment, storage):
    instance_uid = workspace_uid
    machine_type_uid = machine_type
    instance_template_uid = environment
    volume_type_uid = storage

    pattern = re.compile('^[a-z0-9][-a-z0-9]{1,23}[a-z0-9]$')
    if not pattern.match(instance_uid):
        click.echo('Name should be 3 to 25 characters long, lower case alphanumeric or \'-\' and must start and end with an alphanumeric character.')
        sys.exit(1)

    new_instance = {
        'uid': instance_uid,
        'machineType': {
            'uid': machine_type_uid
        },
        'volumeType': {
            'uid': volume_type_uid
        },
        'instanceTemplate': {
            'uid': instance_template_uid
        }
    }

    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    response = ctx.obj['vc'].post(post_object=new_instance)

    if response['status_code'] == 200:
        print('New workspace created: {}'.format(response['data']['uid']))
    else:
        print('An error occurred with creating a new workspace.')
        print("Details: {}".format(response['data']))


@workspaces.command('list', help='Show active workspaces in the current project')
@click.pass_context
@login_required
def list_instances(ctx):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    items = vc.list()

    if items is None or len(items) == 0:
        print('No workspaces found.')
        return

    for item in items:
        info = item['machineType']['info']
        item['cpu'] = info['cpu']
        item['gpu'] = info.get('gpu')
        item['ram'] = info['ram']
        item['hdd'] = item['volumeType']['info']['size']
    vc.print_items(items, fields=['uid', 'cpu', 'gpu', 'ram', 'hdd'], field_names=['ID', 'CPU', 'GPU', 'RAM', 'HDD'])


@workspaces.command('terminate', help='Terminate the workspace')
@click.argument(
    'workspace_uid',
    type=str
)
@click.pass_context
@login_required
def terminate_instance(ctx, workspace_uid):
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()
    ctx.obj['vc'].delete(workspace_uid, message_on_success='Workspace terminated')



""" Command line interface for the OnePanel Machine Learning platform

Entry point for command line interface.
"""


import functools
import os

import click
import requests
from configparser import ConfigParser

from onepanel.commands.common import clone, download, push, pull
from onepanel.utilities.connection import Connection


from onepanel.commands.datasets import datasets
from onepanel.commands.environments import environments
from onepanel.commands.instances import workspaces
from onepanel.commands.jobs import jobs
from onepanel.commands.login import login, login_with_token
from onepanel.commands.machine_types import machine_types
from onepanel.commands.projects import projects
from onepanel.commands.volume_types import volume_types
from onepanel.constants import *
from onepanel.utilities.git_hook_utility import GitHookUtility
from onepanel.utilities.git_utility import CheckGitLfsInstalled


class Connection:
    """ REST API requests defaults and credentials
    """
    def __init__(self):
        self.URL = os.getenv('BASE_API_URL', 'https://c.onepanel.io/api')
        self.SSL_VERIFY = True
        self.headers = {'Content-Type': 'application/json'}
        self.account_uid = None
        self.user_uid = None
        self.token = None
        self.gitlab_impersonation_token = None

        # wrap requests methods to reduce number of arguments in api queries
        self.get = functools.partial(requests.get, headers=self.headers, verify=self.SSL_VERIFY)
        self.post = functools.partial(requests.post, headers=self.headers, verify=self.SSL_VERIFY)
        self.put = functools.partial(requests.put, headers=self.headers, verify=self.SSL_VERIFY)
        self.delete = functools.partial(requests.delete, headers=self.headers, verify=self.SSL_VERIFY)
        self.head = functools.partial(requests.head, headers=self.headers, verify=self.SSL_VERIFY)

    def save_credentials(self, data):
        credentials = ConfigParser()
        credentials['default'] = {'uid': data['uid'],
                                  'token': data['sessions'][0]['token'],
                                  'gitlab_impersonation_token': data['gitlab_impersonation_token'],
                                  'account_uid': data['account']['uid']}

        onepanel_home = os.path.expanduser(os.path.join('~', '.onepanel'))
        if not os.path.exists(onepanel_home):
            os.makedirs(onepanel_home)

        filename = os.path.join(onepanel_home, 'credentials')
        with open(filename, 'w') as f:
            credentials.write(f)

    def load_credentials(self):
        credentials = ConfigParser()
        filename = os.path.expanduser(os.path.join('~', '.onepanel', 'credentials'))
        credentials.read(filename)

        self.user_uid = credentials.get('default','uid',fallback=None)
        self.account_uid = credentials.get('default', 'account_uid', fallback=None)
        self.token = credentials.get('default', 'token', fallback=None)
        self.gitlab_impersonation_token = credentials.get('default','gitlab_impersonation_token', fallback=None)

        if self.token:
            self.headers['Authorization'] = 'Bearer {}'.format(self.token)


@click.group()
@click.version_option(version=CLI_VERSION, prog_name='Onepanel CLI')
@click.pass_context
def cli(ctx):
    conn = Connection()
    conn.load_credentials()
    utility = CheckGitLfsInstalled()

    ctx.obj['connection'] = conn
    ctx.obj['git_utility'] = utility
    utility.figure_out_git_installed()
    if utility.git_installed is False:
        print('Error. Cannot detect git, please verify git is installed.')
        exit(-1)

# Ensure that our git-hook to ping an end-point exits
# Are we in a OnePanel dataset (or project), and is the .git folder available in the current dir?
cwd = os.getcwd()
git_hook_utility = GitHookUtility()
if git_hook_utility.safe_to_add_git_hooks(cwd):
    if git_hook_utility.check_pre_push_gitlab_update_hook(cwd) is False:
        git_hook_utility.add_pre_push_gitlab_update_hook(cwd)
    if git_hook_utility.check_pre_commit_hook(cwd) is False:
        git_hook_utility.add_pre_commit_hook(cwd)

cli.add_command(login)
cli.add_command(login_with_token)
cli.add_command(clone)
cli.add_command(download)
cli.add_command(push)
cli.add_command(pull)
cli.add_command(projects)
cli.add_command(datasets)
cli.add_command(jobs)
cli.add_command(machine_types)
cli.add_command(environments)
cli.add_command(volume_types)
cli.add_command(workspaces)


def main():
    return cli(obj={})


if __name__ == '__main__':
    cli(obj={})



import os, sys, stat
import onepanel.git_hooks.pre_push as pre_push
import onepanel.git_hooks.pre_commit


class GitHookUtility:

    version = None

    def __init__(self):
        self.version = "1.0.0"


    @staticmethod
    def safe_to_add_git_hooks(cwd):
        # Are we in a OnePanel dataset (or project), and is the .git folder available in the current dir?
        onepanel_dir = os.path.join(cwd, '.onepanel')
        git_dir = os.path.join(cwd, '.git')
        if os.path.exists(onepanel_dir) and os.path.exists(git_dir):
            return True
        return False


    def check_pre_push_gitlab_update_hook(self, cwd):
        # Open up the "pre-push" git-hook and verify that it's our git-hook
        git_pre_push_file = os.path.join(cwd,'.git','hooks','pre-push')
        if os.path.isfile(git_pre_push_file):
            with open(git_pre_push_file,'r') as f:
                # First line should be magic shebang with python in it
                first_line_ok = False
                # Second line is an identifier for this code check
                second_line_ok = False
                first_line = f.readline()
                if "python" in first_line:
                    first_line_ok = True
                second_line = f.readline()
                if "onepanel-hook" in second_line:
                    second_line_ok = True
                if first_line_ok and second_line_ok:
                    return True
        return False

    def add_pre_push_gitlab_update_hook(self, cwd):
        # Open up the "pre-push" git-hook and write our hook into it
        git_pre_push_file = os.path.join(cwd, '.git', 'hooks', 'pre-push')
        hook_itself = pre_push.PrePush.read_self()

        with open(git_pre_push_file,'w') as f:
            # Write the first line, which should be the python executable path
            f.write("#!"+sys.executable+'\n')
            # Second line should be our identifier
            f.write("#onepanel-hook"+'\n')
            # The rest of the file should be the hook code
            f.write(hook_itself)
            # ensure file is executable
            st = os.stat(git_pre_push_file)
            os.chmod(git_pre_push_file,st.st_mode | stat.S_IEXEC)

    @staticmethod
    def check_pre_commit_hook(cwd):
        # Open up the "pre-commit" git-hook and verify that it's our git-hook
        git_pre_commit_file = os.path.join(cwd, '.git', 'hooks', 'pre-commit')
        if os.path.isfile(git_pre_commit_file):
            with open(git_pre_commit_file, 'r') as f:
                # First line should be magic shebang with python in it
                first_line_ok = False
                # Second line is an identifier for this code check
                second_line_ok = False
                first_line = f.readline()
                if "python" in first_line:
                    first_line_ok = True
                second_line = f.readline()
                if "onepanel-hook" in second_line:
                    second_line_ok = True
                if first_line_ok and second_line_ok:
                    return True
        return False

    @staticmethod
    def add_pre_commit_hook(cwd):
        # Open up the "pre-commit" git-hook and write our hook into it
        git_pre_commit_file = os.path.join(cwd, '.git', 'hooks', 'pre-commit')
        hook_itself = onepanel.git_hooks.pre_commit.read_self()

        with open(git_pre_commit_file, 'w') as f:
            # Write the first line, which should be the python executable path
            f.write("#!" + sys.executable + '\n')
            # Second line should be our identifier
            f.write("#onepanel-hook" + '\n')
            # The rest of the file should be the hook code
            f.write(hook_itself)
            # ensure file is executable
            st = os.stat(git_pre_commit_file)
            os.chmod(git_pre_commit_file, st.st_mode | stat.S_IEXEC)



"""
This is a git commit-hook which can be used to check if huge files
 where accidentally added to the staging area and are about to be
 committed.
If there is a file which is bigger then the given "max_file_size"-
 variable, the script will exit non-zero and abort the commit.
This script is meant to be added as a "pre-commit"-hook. See this
 page for further information:
    http://progit.org/book/ch7-3.html#installing_a_hook
In order to make the script work probably, you'll need to set the
 above path to the python interpreter (first line of the file)
 according to your system (under *NIX do "which python" to find out).
Also, the "git_binary_path"-variable should contain the absolute
 path to your "git"-executable (you can use "which" here, too).
See the included README-file for further information.
The script was developed and has been confirmed to work under
 python 3.2.2 and git 1.7.7.1 (might also work with earlier versions!)
"""
import subprocess, sys, os
import collections, operator

# The maximum file-size for a file to be committed:
max_file_size = 5120  # in KiB (= 1024 byte)
max_repo_size = 200 * 1024  # in KiB, 200 MB
units = 1024.0  # To move up and down between magnitudes of size


def check_for_large_files():
    # Now, do the checking:
    try:
        too_large_files = []
        # In Bytes
        total_size_of_commit = 0
        LargeFile = collections.namedtuple('LargeFile', ['filename', 'size_bytes'])
        # Check all files in the staging-area:
        text = subprocess.check_output(
            ['git', "status", "--porcelain", "-uno"],
            stderr=subprocess.STDOUT).decode("utf-8")
        file_list = text.splitlines()

        # Check all files:
        for file_s in file_list:
            file_git_status = file_s[:3]
            if "D" in file_git_status:
                continue
            file_name = file_s[3:]
            stat = os.stat(file_name)
            total_size_of_commit += stat.st_size
            if stat.st_size >= (max_file_size * units):
                # File is too big, store it for later analysis
                too_large_files.append(LargeFile(file_name, stat.st_size))

        # Are they trying to commit large files?
        if len(too_large_files) > 0:
            print("The total size of your commits are too large.")
            print("Use Onepanel datasets to store datasets or any files with combined size larger than 5.0 MB.")
            print("Use \"git reset HEAD <file>\" to unstage large files and try your commit.")
            num_files_to_show = 5
            print("Top {to_show} largest files...".format(to_show=num_files_to_show))
            print_top_x_large_files(num_files_to_show, too_large_files)
            sys.exit(1)

        # Is the total commit size greater than 5 mb?
        # This is to prevent thousands of small image files from being committed.
        if total_size_of_commit >= (max_file_size * units):
            print("The total size of your commit is too large, based on the total size of the files being committed.")
            print("Please use a dataset instead of git to store these files.")
            print("Adjust your commit and try again.")
            print("Note that you may visit https://help.onepanel.io/command-line-interface-cli/getting-started-with"
                  "-onepanel-cli for more information.")
            sys.exit(1)

        # Check if the repository size is too big
        git_repo_info = subprocess.check_output(
            ['git', "count-objects", "-v"],
            stderr=subprocess.STDOUT).decode("utf-8")
        repo_info_output = git_repo_info.splitlines()
        size_line = repo_info_output[1]
        kib_info = size_line.split(":")
        kib_val = int(kib_info[1].strip())
        if kib_val * units >= max_repo_size * units:
            print("The total size of your repository is too large, greater than or equal to 200 MB.")
            print("Please use a dataset instead of git to store these files.")
            print("Note that you may visit https://help.onepanel.io/command-line-interface-cli/getting-started-with"
                  "-onepanel-cli for more information.")
            sys.exit(1)

        # Everything seems to be okay with file sizes
        sys.exit(0)

    except subprocess.CalledProcessError:
        # There was a problem calling "git status".
        print("Error executing git status")
        sys.exit(12)


def sizeof_fmt(num):
    """
	This function will return a human-readable filesize-string
	 like "3.5 MB" for it's given 'num'-parameter.
	From http://stackoverflow.com/questions/1094841
	"""
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < units:
            return "%3.1f %s" % (num, x)
        num /= units


def print_top_x_large_files(how_many=5, list_of_files=None):
    if list_of_files is None:
        list_of_files = []
    # Sort the list of files by size
    sorted_files = sorted(list_of_files, key=operator.attrgetter('size_bytes'), reverse=True)
    for index, file in enumerate(sorted_files):
        if index == how_many:
            break
        print(
            "{file_name} - ({size_of_file})".format(file_name=file.filename, size_of_file=sizeof_fmt(file.size_bytes)))



import subprocess


class CheckGitLfsInstalled:
    git_major_vr = 0
    git_sub_vr = 0
    git_minor_vr = 0

    def __init__(self):
        self.git_installed = False

    def figure_out_git_installed(self):
        shell_cmd = ['git','version']
        try:
            p = subprocess.Popen(shell_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 shell=False)
            p.wait()
            line = p.stdout.readline()
            git_output = line.decode().rstrip()
            if "git version" in git_output:
                self.git_installed = True
                # Figure out the git version installed
                self.parse_git_version(git_output)
        except FileNotFoundError:
            print("An error occurred when trying to figure out if git is installed.")

    def parse_git_version(self, shell_str):
        split_list = shell_str.split(" ")
        if len(split_list) == 3:
            git_version = split_list[2]
            git_version_list = git_version.split(".")
            # Example: "2.17.0"
            if len(git_version_list) >= 3:
                self.git_major_vr = int(git_version_list[0])
                self.git_sub_vr = int(git_version_list[1])
                self.git_minor_vr = int(git_version_list[2])

    def get_git_clone_str(self):
        return 'git clone'


import os

from onepanel.utilities.connection import Connection
from onepanel.commands.datasets import DatasetViewController
from onepanel.commands.projects import ProjectViewController


class PrePushToEndpointHook:

    version = None

    def __init__(self):
        self.version = "1.0.0"


    @staticmethod
    def ping_endpoint():
        conn = Connection()
        conn.load_credentials()
        data = {
            'version':'1',
            'data':''
        }

        # Figure out if this is a project or dataset push
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot connect to endpoint.")
            print("DETAILS." + onepanel_dir)
            exit(-1)

        project_push = True
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        if not os.path.isfile(project_file):
            project_push = False
        else:
            # Safe to ping the endpoint with project credentials
            pvc = ProjectViewController(conn)
            pvc.init_credentials_retrieval()
            url_params = '/gitlab/accounts/{account_uid}/project/{project_uid}/update'.format(
                account_uid=pvc.account_uid,project_uid=pvc.project_uid
            )
            try:
                pvc.post(data,params=url_params)
            except ValueError as e:
                print("Gitlab Update Failed.")

        if project_push == False:
            print("ERROR. Cannot connect to endpoint, unclear if this code is initialized with 'onepanel init'.")
            exit(-1)



""" Command line interface for the OnePanel Machine Learning platform

'Datasets' commands group.
"""
import glob
import os
import re

import click
import configobj
import humanize

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.utilities.aws_utility import AWSUtility
from onepanel.utilities.s3_authenticator import S3Authenticator


class DatasetViewController(APIViewController):
    """ DatasetViewController data model
    """

    DATASET_FILE = os.path.join('.onepanel','dataset')
    EXCLUSIONS = [os.path.join('.onepanel','dataset')]

    account_uid = None
    dataset_uid = None

    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.endpoint = '{}'.format(
            self.conn.URL,
        )

    def init_credentials_retrieval(self):
        # Figure out the account uid and dataset uid
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all datasets operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)
        if not os.path.isfile(dataset_file):
            print("ERROR.Dataset file does not exist, cannot carry out all datasets operations.")
            print("DETAILS." + dataset_file)
            exit(-1)

        cfg = configobj.ConfigObj(dataset_file)

        dataset_uid = cfg['uid']
        dataset_account_uid = cfg['account_uid']

        if len(dataset_uid) < 1 or len(dataset_account_uid) < 1:
            print("ERROR.Dataset file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.account_uid = dataset_account_uid
        self.dataset_uid = dataset_uid


    def save(self, home):
        if not os.path.exists(home):
            os.makedirs(home)
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            os.makedirs(onepanel_dir)
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)

        cfg = configobj.ConfigObj(dataset_file)
        cfg['uid'] = self.dataset_uid
        cfg['account_uid'] = self.account_uid
        cfg.write()

    def init_endpoint(self):
        self.endpoint = '{root}/accounts/{account_uid}/datasets/{dataset_uid}'.format(
            root=self.conn.URL,
            account_uid=self.account_uid,
            dataset_uid=self.dataset_uid
        )

    @classmethod
    def from_json(cls, data):
        cls.account_uid = data['account']['uid']
        cls.dataset_uid = data['uid']

    @classmethod
    def is_uid_valid(cls,uid):
        pattern = re.compile('^[a-z0-9][-a-z0-9]{1,23}[a-z0-9]$')
        if pattern.match(uid):
            return True
        else:
            return False

    @classmethod
    def exists_local(cls,home):
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)
        if os.path.isfile(dataset_file):
            return True
        else:
            return False

    @classmethod
    def exists_remote(cls,dataset_uid, data):
        exists = False
        if data['uid'] == dataset_uid:
            exists = True
        return exists


def create_dataset(ctx, account_uid, home):
    """ Dataset creation method for 'datasets_init' and 'datasets_create'
    commands
    """
    vc = ctx.obj['vc']

    if not account_uid:
        account_uid = ctx.obj['connection'].account_uid

    dataset_uid = os.path.basename(home)
    if not vc.is_uid_valid(dataset_uid):
        click.echo('Dataset name {} is invalid.'.format(dataset_uid))
        click.echo(
            'Name should be 3 to 25 characters long, lower case alphanumeric or \'-\' and must start and end with an alphanumeric character.')
        return None

    # Check if the dataset already exists remotely
    url_dataset_check = '/accounts/{}/datasets/{}'.format(account_uid, dataset_uid)
    response_data = vc.get(params=url_dataset_check)
    remote_dataset = response_data['data']

    dataset = None
    # If remote returned data, do operations
    if remote_dataset is not None and vc.exists_remote(dataset_uid, remote_dataset):
        click.echo("Dataset already exists. Please download the dataset if you want to use it.")
        return None
    # If no data from remote, we can create a dataset
    else:
        can_create = True
        if vc.exists_local(home):
            can_create = click.confirm(
                'Dataset exists locally but does not exist in {}, create the dataset and remap local folder?'
                    .format(account_uid))

        if can_create:
            data = {
                'uid': dataset_uid
            }
            url_dataset_create = '/accounts/{}/datasets'.format(account_uid)
            response = vc.post(data,params=url_dataset_create)
            if response['status_code'] == 200:
                vc.from_json(response['data'])
                vc.save(home)
            else:
                click.echo("Encountered error.")
                click.echo(response['status_code'])
                click.echo(response['data'])
                return False
    return True


@click.group(help='Dataset commands group')
@click.pass_context
def datasets(ctx):
    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])


@datasets.command('list', help='Display a list of all datasets.')
@click.pass_context
@login_required
def datasets_list(ctx):
    vc = ctx.obj['vc']
    data = vc.list(params='/datasets')
    if data is None or len(data['data']) < 1:
        print("No datasets found.")
    else:
        data_print = []
        for datum in data['data']:
            uid = datum['uid']
            version_count = datum['statistics']['version']
            statistics_data = datum['statistics']['data']
            size = 0
            if len(statistics_data) > 0:
                size = datum['statistics']['data']['storage_size']
            size_formatted = humanize.naturalsize(size, binary=True)
            data_print.append({'uid': uid,
                               'version_count': version_count,
                               'size': size_formatted
                               })

        empty_message = "No datasets found."
        fields = ['uid', 'version_count', 'size']
        field_names = ['NAME', 'VERSIONS', 'SIZE']
        DatasetViewController.print_items(data_print, fields, field_names, empty_message)


@datasets.command('init', help='Initialize dataset in current directory.')
@click.pass_context
@login_required
def datasets_init(ctx):
    vc = ctx.obj['vc']
    home = os.getcwd()
    if not vc.is_uid_valid(os.path.basename(home)):
        dataset_uid = click.prompt('Please enter a valid dataset name')
        home = os.path.join(home, dataset_uid)

    if create_dataset(ctx, None, home):
        click.echo('Dataset is initialized in current directory.')


@datasets.command('create', help='Create dataset in new directory.')
@click.argument('name', type=str)
@click.pass_context
@login_required
def datasets_create(ctx, name):
    home = os.path.join(os.getcwd(), name)
    if create_dataset(ctx, None, home):
        click.echo('Dataset is created in directory {}.'.format(home))

@datasets.command('push', help='Push up dataset changes')
@click.option(
    '-m', '--message',
    type=str,
    default=None,
    help='Datasets only: Add a message to this version. Up to 255 chars.\"text\".'
)
@click.option(
    '-n', '--name',
    type=str,
    default=None,
    help='Datasets only: Add a name to this version. Use \"text\".'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.pass_context
@login_required
def datasets_push(ctx,message, name,quiet):
   general_push(ctx,message, name,quiet)

def datasets_clone(ctx, path, directory, suppress_output=False):
    conn = ctx.obj['connection']
    vc = DatasetViewController(conn)

    values = path.split('/')

    if len(values) == 3:
        try:
            account_uid, datasets_dir, dataset_uid = values
            assert (datasets_dir == 'datasets')
        except:
            click.echo('Invalid dataset path. Please use <account_uid>/datasets/<uid>')
            return
    else:
        click.echo('Invalid dataset path. Please use <account_uid>/datasets/<uid>')
        return

    vc.account_uid = account_uid
    vc.dataset_uid = dataset_uid

    # check dataset path, account_uid, dataset_uid
    if directory is None:
        home = os.path.join(os.getcwd(), dataset_uid)
    elif directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    # check if the dataset exists
    url_dataset_check = '/accounts/{}/datasets/{}'.format(account_uid, dataset_uid)
    response_data = vc.get(uid='', field_path='',params=url_dataset_check)
    response_code = response_data['status_code']
    if response_code == 200:
        remote_dataset = response_data['data']
    elif response_code == 401 or response_code == 404:
        print('Dataset does not exist.')
        return
    else:
        print('Error: {}'.format(response_code))
        return None

    if not vc.exists_remote(dataset_uid, remote_dataset):
        click.echo('There is no dataset {}/datasets/{} on the server'.format(account_uid, dataset_uid))
        return

    can_create = True
    if vc.exists_local(home):
        can_create = click.confirm('Dataset already exists, overwrite?')
    if not can_create:
        return
    # Authenticate for S3
    s3auth = S3Authenticator(ctx.obj['connection'])
    creds = s3auth.get_s3_credentials(vc.account_uid, 'datasets', vc.dataset_uid)
    if creds['data'] is None:
        print("Unable to get S3 credentials. Exiting.")
        return False
    aws_access_key_id = creds['data']['AccessKeyID']
    aws_secret_access_key = creds['data']['SecretAccessKey']
    aws_session_token = creds['data']['SessionToken']
    aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                          aws_secret_access_key=aws_secret_access_key,
                          aws_session_token=aws_session_token)
    aws_util.suppress_output = suppress_output
    vc.init_endpoint()

    current_version = vc.get(field_path='/version/current')

    if current_version['data'] is None:
        if current_version['status_code'] == 404:
            click.echo('Dataset is without files.')
            vc.save(home)
            return 0
        else:
            click.echo('Could not get information about dataset.')
            return

    if current_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    vc.save(home)
    s3_from_dir = current_version['data']['version']['path']
    exit_code = aws_util.download_all(home, s3_from_dir)
    if exit_code != 0:
        click.echo('\nError with downloading files.')
        return

    return 0


def general_push(ctx,comment='',name='',suppress_output=False):
    files = glob.glob("*")
    if len(files) < 1:
        click.echo("Cannot find any files in current dir, exiting.")
        return

    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    # Before trying to do any upload, make sure user has permission
    # For example, we don't want them to modify a public dataset
    url_dataset_member_check = '/is_member'
    response_data = vc.head(uid='', field_path='', params=url_dataset_member_check)
    if response_data['status_code'] == 404:
        click.echo('You are not authorized to push to this dataset.')
        return

    next_version = vc.get(field_path='/version/next')

    if next_version['data'] is None:
        click.echo('Could not get information about dataset.')
        return

    if next_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    # Authenticate for S3
    s3auth = S3Authenticator(ctx.obj['connection'])
    creds = s3auth.get_s3_credentials(vc.account_uid, 'datasets', vc.dataset_uid)
    if creds['data'] is None:
        print("Unable to get S3 credentials. Exiting.")
        return False
    aws_access_key_id = creds['data']['AccessKeyID']
    aws_secret_access_key = creds['data']['SecretAccessKey']
    aws_session_token = creds['data']['SessionToken']

    dataset_dir = os.curdir
    s3_push_to_dir = next_version['data']['version']['path']
    aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                          aws_secret_access_key=aws_secret_access_key,
                          aws_session_token=aws_session_token)
    aws_util.suppress_output = suppress_output
    exit_code = aws_util.upload_dir(dataset_dir, s3_push_to_dir, '.onepanel/*')
    if exit_code != 0:
        click.echo('\nError with pushing up files.')
        return

    # Add the users' 'name' and 'comment' to the dataset
    curr_version = next_version['data']['version']['version']
    post_obj = {
        'comment': comment,
        'name': name
    }
    update_dataset_resp = vc.put(field_path='/update_version',post_object=post_obj)
    if update_dataset_resp['status_code'] != 200:
        click.echo('\nError with calling the API. Contact support if this error continues.')
        return

    click.echo('\nPushed up version ' + str(curr_version))


def datasets_download(ctx, path, directory,suppress_output=False):
    if directory is None or directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    code = datasets_clone(ctx, path, home,suppress_output=suppress_output)
    if code != 0:
        print("Unable to download!")
        return False

    print('The files have been downloaded to: {dir}'.format(dir=home))
    return True



from setuptools import setup

from onepanel.constants import *

setup(
    name="onepanel",
    version=CLI_VERSION,
    packages = ['onepanel', 'onepanel.commands','onepanel.utilities','onepanel.git_hooks'],
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    install_requires=[
        'prettytable',
        'requests',
        'click',
        'PTable',
        'configobj',
        'websocket-client',
        'humanize',
        'configparser',
        'awscli'
    ],
    entry_points='''
        [console_scripts]
        onepanel=onepanel.cli:main
    ''',
)



import os

import click

from onepanel.commands import jobs
from onepanel.commands.datasets import datasets_clone, datasets_download, DatasetViewController, general_push
from onepanel.commands.jobs import jobs_download_output, JobViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import projects_clone, ProjectViewController
from onepanel.gitwrapper import GitWrapper


def get_entity_type(path):
    dirs = path.split('/')
    entity_type = None
    if len(dirs) == 3:
        account_uid, dir, uid = dirs
        if dir == 'projects':
            entity_type = 'project'
        elif dir == 'datasets':
            entity_type = 'dataset'
    elif len(dirs) == 5:
        account_uid, parent, project_uid, dir, uid = dirs
        if parent == 'projects' and dir == 'jobs':
            entity_type = 'job'
    return entity_type

@click.command('clone', help='Clone project or dataset from server.')
@click.argument('path', type=click.Path())
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.pass_context
@login_required
def clone(ctx, path, directory,quiet):
    entity_type = get_entity_type(path)
    if entity_type == 'project':
        projects_clone(ctx, path, directory)
    elif entity_type == 'dataset':
        datasets_clone(ctx, path, directory,quiet)
    else:
        click.echo('Invalid project or dataset path.')

@click.command('download', help='Download a dataset')
@click.argument('path', type=click.Path())
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '--archive',
    type=bool,
    is_flag=True,
    default=False,
    help='Download the output as a compressed file.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.pass_context
@login_required
def download(ctx, path, directory,archive,quiet):
    entity_type = get_entity_type(path)
    if entity_type == 'dataset':
        datasets_download(ctx, path, directory,quiet)
    elif entity_type == 'job':
        jobs_download_output(ctx, path, directory,archive)
    else:
        click.echo('Invalid path.')

@click.command('push', help='Push changes to onepanel')
@click.option(
    '-m', '--message',
    type=str,
    default=None,
    help='Datasets only: Add a message to this version. Up to 255 chars.\"text\".'
)
@click.option(
    '-n', '--name',
    type=str,
    default=None,
    help='Datasets only: Add a name to this version. Use \"text\".'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.pass_context
@login_required
def push(ctx, message, name,quiet):
    home = os.getcwd()
    # Are we uploading job output? A project? Or Dataset?
    if os.path.isfile(JobViewController.JOB_OUTPUT_FILE):
        jobs.upload_output(ctx,quiet)
    elif os.path.isfile(ProjectViewController.PROJECT_FILE):
        if message is not None or name is not None:
            click.echo(
                "Projects do not support these arguments. Remove arguments and try again.")
            return
        GitWrapper().push(home)
    elif os.path.isfile(DatasetViewController.DATASET_FILE):
        general_push(ctx, message, name,quiet)
    else:
        click.echo("Cannot determine if you are trying to push job output, dataset files, or project files. Are you in the right directory?")

@click.command('pull', help='Pull changes from onepanel (fetch and merge)')
@click.pass_context
@login_required
def pull(ctx):
    home = os.getcwd()
    GitWrapper().pull(home)
